from accern import AccernClient, error
import pytest


def test_fails_without_token():
    with pytest.raises(error.AccernError) as exc_info:
        AccernClient.check_token(token=None)
    assert exc_info.value.args[0] == 'No token provided.'

def test_build_api_param_no_harvested_at():
    schema = {
        'filters': {
            'harvested_at': ['2017-12-26 00:00:00', '2017-12-27 00:00:00']
        }
    }
    resp = AccernClient.build_api_params(schema)
    assert 'harvested_at' not in resp and 'from' in resp



from datetime import datetime
import logging
import os
import re
import six
import sys
try:
    from urlparse import urlsplit, urlunsplit
except ImportError:
    from urllib.parse import urlsplit, urlunsplit
try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode

ACCERN_LOG = os.environ.get('ACCERN_LOG')
logger = logging.getLogger('accern')

__all__ = [
    'datetime',
    'json',
    'log_debug',
    'log_info',
    'logfmt',
    'six',
    'urlencode',
    'urlsplit',
    'urlunsplit'
]

try:
    import json
except ImportError:
    json = None

if not (json and hasattr(json, 'loads')):
    try:
        import simplejson as json
    except ImportError:
        if not json:
            raise ImportError(
                "Accern requires a JSON library, such as simplejson. "
                "HINT: Try installing the "
                "python simplejson library via 'pip install simplejson' or "
                "'easy_install simplejson', or contact support@accern.com "
                "with questions.")
        else:
            raise ImportError(
                "Accern requires a JSON library with the same interface as "
                "the Python 2.6 'json' library.  You appear to have a 'json' "
                "library with a different interface.  Please install "
                "the simplejson library.  HINT: Try installing the "
                "python simplejson library via 'pip install simplejson' "
                "or 'easy_install simplejson', or contact support@accern.com"
                "with questions.")


def utf8(value):
    if six.PY2 and isinstance(value, unicode):
        return value.encode('utf-8')
    return value


def _console_log_level():
    if ACCERN_LOG in ['debug', 'info']:
        return ACCERN_LOG
    return None


def log_debug(message, **params):
    msg = logfmt(dict(message=message, **params))
    if _console_log_level() == 'debug':
        print(msg, sys.stderr)
    logger.debug(msg)


def log_info(message, **params):
    msg = logfmt(dict(message=message, **params))
    if _console_log_level() in ['debug', 'info']:
        print(msg, sys.stderr)
    logger.info(msg)


def logfmt(props):
    def fmt(key, val):
        # Handle case where val is a bytes or bytesarray
        if six.PY3 and hasattr(val, 'decode'):
            val = val.decode('utf-8')
        # Check if val is already a string to avoid re-encoding into
        # ascii. Since the code is sent through 2to3, we can't just
        # use unicode(val, encoding='utf8') since it will be
        # translated incorrectly.
        if not isinstance(val, six.string_types):
            val = six.text_type(val)
        if re.search(r'\s', val):
            val = repr(val)
        # key should already be a string
        if re.search(r'\s', key):
            key = repr(key)
        return u'{key}={val}'.format(key=key, val=val)
    return u' '.join([fmt(key, val) for key, val in sorted(props.items())])



"""Python library for Accern API.

Core client functionality, common across all API requests (including performing
HTTP requests).
"""

from accern import default_client, error, util
from accern.schema import Schema
from accern.default_client import AccernClient

API_BASE = "https://admin-staging.accern.com/api/io/jobs"


class HistoricalClient(AccernClient):
    """Perform requests to the Accern API web services."""

    def __init__(self, token=None, client=None):
        """Intialize with params.

        :param client: default http client. Optional
        :param token: Accern API token. Required.
        """
        self.api_base = API_BASE
        self.token = token
        self._client = client or default_client.new_http_client()

    @staticmethod
    def interpret_response(rbody, rcode, rheaders):
        try:
            if hasattr(rbody, 'decode'):
                rbody = rbody.decode('utf-8')
            resp = util.json.loads(rbody)
        except Exception:
            raise error.APIError(
                "Invalid response body from API: %s "
                "(HTTP response code was %d)" % (rbody, rcode),
                rbody, rcode, rheaders)

        if not 200 <= rcode < 300:
            raise AccernClient.handle_error(rbody, rcode, resp)

        return resp

    def create_job(self, schema):
        """Create a job with schema.

        :param schema: job detail, will be added to payload

        :raises ApiError: when the API returns an error.
        :raises Timeout: if the request timed out.
        :raises TransportError: when something went wrong while trying to
            exceute a request.
        """

        schema = Schema.validate_schema(method='historical', schema=schema)
        token = AccernClient.check_token(self.token)
        method = 'POST'

        headers = AccernClient.build_api_headers(token, method)

        if method == 'POST':
            post_data = util.json.dumps({'query': schema})
        else:
            post_data = None

        rbody, rcode, rheaders = self._client.request(method, self.api_base, headers, post_data)
        resp = self.interpret_response(rbody, rcode, rheaders)
        return resp

    def get_jobs(self, job_id=None):
        """Get the user's job history.

        :param job_id: if job_id is valid, will return the job related
        """

        token = AccernClient.check_token(self.token)
        method = 'GET'
        headers = AccernClient.build_api_headers(token, method)
        if job_id is None:
            rbody, rcode, rheaders = self._client.request(method, self.api_base, headers, post_data=None)
            resp = self.interpret_response(rbody, rcode, rheaders)
        else:
            rbody, rcode, rheaders = self._client.request(method, '%s/%s' % (self.api_base, job_id), headers, post_data=None)
            resp = self.interpret_response(rbody, rcode, rheaders)

        return resp



"""Accern REST APIs.

Core client functionality, common across all API requests (including performing
HTTP requests).
"""

from accern import error, util
from accern.schema import Schema
from accern.default_client import AccernClient, new_http_client
API_BASE = "https://feed.accern.com/v4/alphas"


class API(AccernClient):
    """Perform requests to the Accern API web services."""

    def __init__(self, token=None, client=None):
        """Intialize with params.

        :param client: default http client. Optional
        :param token: Accern API token. Required.
        """
        self.api_base = API_BASE
        self.token = token
        self._client = client or new_http_client()

    @staticmethod
    def interpret_response(rbody, rcode, rheaders, schema):
        try:
            if hasattr(rbody, 'decode'):
                rbody = rbody.decode('utf-8')
            resp = util.json.loads(rbody)
        except Exception:
            raise error.APIError(
                "Invalid response body from API: %s "
                "(HTTP response code was %d)" % (rbody, rcode),
                rbody, rcode, rheaders)

        if not 200 <= rcode < 300:
            raise error.APIError('API request failed.')

        if resp['total'] > 0:
            resp['signals'] = AccernClient.quant_filter(schema, resp['signals'])
            resp['signals'] = AccernClient.select_fields(schema, resp['signals'])
            resp['total'] = len(resp['signals'])
        return resp

    def request(self, schema=None):
        rbody, rcode, rheaders = self.request_raw(schema)
        resp = self.interpret_response(rbody, rcode, rheaders, schema)
        return resp

    def request_raw(self, schema):
        """Perform HTTP GET with credentials.

        :raises AuthenticationError: when the token is invalid.
        """
        schema = Schema.validate_schema(method='api', schema=schema)

        params = AccernClient.build_api_params(schema)
        params['token'] = AccernClient.check_token(self.token)
        encoded_params = util.urlencode(list(AccernClient.api_encode(params)))
        abs_url = AccernClient.build_api_url(self.api_base, encoded_params)

        rbody, rcode, rheaders = self._client.request('GET', abs_url, headers=None, post_data=None)
        return rbody, rcode, rheaders



from accern import Schema, error
import pytest


def test_fails_get_options_with_invalid_field_name():
    with pytest.raises(error.SchemaError) as exc_info:
        Schema.get_options(field='test')
    assert exc_info.value.args[0] == 'Invalid field (test) in filter option.'


def test_get_url_params():
    resp = Schema.get_url_params()
    assert len(resp) == 15

############################################################
# Test Cases for func: validate_options                    #
############################################################


def test_fails_with_invalid_field():
    with pytest.raises(error.SchemaError) as exc_info:
        Schema.validate_options(field='test')
    assert exc_info.value.args[0] == 'Invalid field (test) in filter option.'


def test_fails_without_value():
    with pytest.raises(error.SchemaError) as exc_info:
        Schema.validate_options(field='entity_ticker')
    assert exc_info.value.args[0] == 'No filter option value for "entity_ticker".'


def test_fails_range_malformed_value():
    resp = Schema.validate_options(field='entity_sentiment', value=[0, -5])
    assert resp['error'] == '"entity_sentiment" has malformed filter option value.'


def test_fails_range_with_more_argument_value():
    resp = Schema.validate_options(field='entity_sentiment', value=[0, -5, 11])
    assert resp['error'] == '"entity_sentiment" has wrong number or arguments.'


def test_fails_range_with_less_argument_value():
    resp = Schema.validate_options(field='entity_sentiment', value=[11])
    assert resp['error'] == '"entity_sentiment" has wrong number or arguments.'


def test_fails_no_range_malformed_value():
    resp = Schema.validate_options(field='story_traffic', value=[0, -5])
    assert resp['error'] == '"story_traffic" has malformed filter option value.'


def test_fails_no_range_with_more_argument_value():
    resp = Schema.validate_options(field='story_traffic', value=[0, -5, 11])
    assert resp['error'] == '"story_traffic" has wrong number or arguments.'


def test_fails_no_range_with_less_argument_value():
    resp = Schema.validate_options(field='story_traffic', value=[11])
    assert resp['error'] == '"story_traffic" has wrong number or arguments.'


############################################################
# Test Cases for func: validate_schema                     #
############################################################


def test_fails_without_method():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {}
        Schema.validate_schema(schema=schema)
    assert exc_info.value.args[0] == 'Method is missing.'


def test_with_empty_schema():
    schema = Schema.validate_schema(method='api')
    assert schema is None


def test_uppercase_schema():
    schema = {
        'FILTERS': {
            'entity_sentiment': [0, 10]
        },
        'SELECT': [
            {
                'field': 'entity_sentiment'
            }
        ]
    }
    schema = Schema.validate_schema(method='api', schema=schema)
    assert schema == {
        'filters': {'entity_sentiment': [0, 10]},
        'select': [{'alias': 'entity_sentiment', 'field': 'entity_sentiment'}]
    }

############################################################
# Test cases for func: validate_schema, method: api        #
############################################################


def test_fails_with_less_argument_value_in_filters():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {
            'filters': {
                'entity_sentiment': [
                    [0, 10],
                    [15]
                ]
            }
        }
        Schema.validate_schema(method='api', schema=schema)
    assert exc_info.value.args[0] == '"entity_sentiment" has wrong number or arguments.'


def test_fails_with_name_in_api_schema():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {
            'name': 'test'
        }
        Schema.validate_schema(method='api', schema=schema)
    assert exc_info.value.args[0] == 'Illegal "name" in api schema.'


def test_fails_with_description_in_api_schema():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {
            'description': 'test'
        }
        Schema.validate_schema(method='api', schema=schema)
    assert exc_info.value.args[0] == 'Illegal "description" in api schema.'


def test_fails_with_multiple_filters_in_api_schema():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {
            'filters': [
                {
                    'entity_sentiment': [0, 10]
                },
                {
                    'entity_sentiment': [50, 100]
                }
            ]
        }
        Schema.validate_schema(method='api', schema=schema)
    assert exc_info.value.args[0] == 'Method "api" does not support multiple filters.'


def test_fails_with_select_function_in_api_schema():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {
            'select': [
                {
                    'field': 'harvested_at',
                    'alias': 'month',
                    'function': 'month'
                }
            ]
        }
        Schema.validate_schema(method='api', schema=schema)
    assert exc_info.value.args[0] == 'Method "api" does not support select field functions.'


def test_fails_with_missing_field_in_select_function_in_api_schema():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {
            'select': [
                {
                    'alias': 'harvested_at'
                }
            ]
        }
        Schema.validate_schema(method='api', schema=schema)
    assert exc_info.value.args[0] == 'Missing "field" in select option.'

############################################################
# Test cases for func: validate_schema, method: stream     #
############################################################


def test_fails_with_name_in_stream_schema():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {
            'name': 'test'
        }
        Schema.validate_schema(method='stream', schema=schema)
    assert exc_info.value.args[0] == 'Illegal "name" in stream schema.'


def test_fails_with_description_in_stream_schema():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {
            'description': 'test'
        }
        Schema.validate_schema(method='stream', schema=schema)
    assert exc_info.value.args[0] == 'Illegal "description" in stream schema.'


def test_fails_with_multiple_filters_in_stream_schema():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {
            'filters': [
                {
                    'entity_sentiment': [0, 10]
                },
                {
                    'entity_sentiment': [50, 100]
                }
            ]
        }
        Schema.validate_schema(method='stream', schema=schema)
    assert exc_info.value.args[0] == 'Method "stream" does not support multiple filters.'


def test_fails_with_select_function_in_stream_schema():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {
            'select': [
                {
                    'field': 'harvested_at',
                    'alias': 'month',
                    'function': 'month'
                }
            ]
        }
        Schema.validate_schema(method='stream', schema=schema)
    assert exc_info.value.args[0] == 'Method "stream" does not support select field functions.'


def test_fails_with_missing_field_in_select_function_in_stream_schema():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {
            'select': [
                {
                    'alias': 'harvested_at'
                }
            ]
        }
        Schema.validate_schema(method='stream', schema=schema)
    assert exc_info.value.args[0] == 'Missing "field" in select option.'

############################################################
# Test cases for func: validate_schema, method: historical #
############################################################


def test_fails_with_no_name_in_historical_schema():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {
            'description': 'test'
        }
        Schema.validate_schema(method='historical', schema=schema)
    assert exc_info.value.args[0] == 'Required field "name" not found in historical schema.'


def test_fails_with_no_description_in_historical_schema():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {
            'name': 'test'
        }
        Schema.validate_schema(method='historical', schema=schema)
    assert exc_info.value.args[0] == 'Required field "description" not found in historical schema.'


def test_fails_with_less_argument_value_in_multiple_filters():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {
            'name': 'test',
            'description': 'test',
            'filters': [
                {
                    'entity_sentiment': [0, 10]
                },
                {
                    'entity_sentiment': [0]
                }
            ]
        }
        Schema.validate_schema(method='historical', schema=schema)
    assert exc_info.value.args[0] == '"entity_sentiment" has wrong number or arguments.'


def test_fails_with_missing_field_in_select_function_in_historical_schema():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {
            'name': 'test',
            "description": 'test',
            'select': [
                {
                    'alias': 'harvested_at'
                }
            ]
        }
        Schema.validate_schema(method='historical', schema=schema)
    assert exc_info.value.args[0] == 'Missing "field" in select option.'


def test_fails_alias_diff_function_for_harvested_at():
    with pytest.raises(error.SchemaError) as exc_info:
        schema = {
            'name': 'test',
            'description': 'test',
            'select': [
                {
                    'field': 'harvested_at',
                    'alias': 'hour',
                    'function': 'hourly'
                }
            ]
        }
        Schema.validate_schema(method='historical', schema=schema)
    assert exc_info.value.args[0] == "Alias of harvested_at is different from it's aggregation function."


def test_get_agg_function_for_missing_categorical_field():
    schema = {
        'name': 'test',
        'description': 'test',
        'select': [
            {
                'field': 'entity_ticker'
            },
            {
                'field': 'harvested_at',
                'alias': 'hour',
                'function': 'hour'
            }
        ]
    }
    schema = Schema.validate_schema(method='historical', schema=schema)
    assert schema['select'][0]['function'] == 'group'


def test_get_agg_function_for_missing_range_field():
    schema = {
        'name': 'test',
        'description': 'test',
        'select': [
            {
                'field': 'entity_sentiment'
            },
            {
                'field': 'harvested_at',
                'alias': 'hour',
                'function': 'hour'
            }
        ]
    }
    schema = Schema.validate_schema(method='historical', schema=schema)
    assert schema['select'][0]['function'] == 'average'



from accern import error

import ConfigParser
import os

__all__ = [
    'get_config'
]


def get_config(section):
    Config = ConfigParser.ConfigParser()
    if not os.path.exists('accern.ini'):
        raise error.ConfigError("Config file not find.")

    Config.read('accern.ini')
    dict1 = {}
    options = Config.options(section)
    for option in options:
        try:
            dict1[option] = Config.get(section, option)
            if dict1[option] == -1:
                print("skip: %s" % option)
        except ValueError:
            print("exception on %s!" % option)
            dict1[option] = None
    return dict1



"""Accern Streaming API client.

Core client functionality, common across all API requests (including performing
HTTP requests).
"""
import codecs
import re
import requests
import time
from accern.default_client import AccernClient, Event
from accern.schema import Schema
from accern import util

API_BASE = 'https://feed.accern.com/v4/stream'
END_OF_FIELD = re.compile(r'\r\n\r\n|\r\r|\n\n')


class StreamListener(AccernClient):
    def on_data(self, raw_data):
        """Call when raw data is received from connection.

        Override this method if you want to manually handle stream data. Return
        False to stop stream and close connection.
        """
        if 'disconnect' in raw_data:
            if self.on_disconnect(raw_data['disconnect']) is False:
                return False
        return raw_data

    @staticmethod
    def on_disconnect(notice):
        """Call when server return a disconnect notice."""
        if notice == 'disconnect':
            return False
        return True


class StreamClient(object):
    """Perform requests to the Accern API web services."""

    def __init__(self, listener, token=None, schema=None, **kwargs):
        """Intialize with params.

        :param client: default http client. Optional
        :param token: Accern API token. Required.
        """
        self._listener = listener or StreamListener()
        self.api_base = API_BASE
        # Keep data here as it streams in
        self.buf = u''
        self.chunk_size = kwargs.get('chunk_size', 1024)
        self.last_id = kwargs.get('last_id', None)
        self.schema = schema or {}
        # Any extra kwargs will be fed into the requests.get call later.
        self.requests_kwargs = kwargs.get('request', {})
        self.resp = None
        self.resp_iterator = None
        self.retry = kwargs.get('retry', 3000)
        self.timeout = kwargs.get('timeout', 300.0)
        self.token = token or None
        self.url = None

        # The SSE spec requires making requests with Cache-Control: nocache
        if 'headers' not in self.requests_kwargs:
            self.requests_kwargs['headers'] = {}
        self.requests_kwargs['headers']['Cache-Control'] = 'no-cache'

        # The 'Accept' header is not required, but explicit > implicit
        self.requests_kwargs['headers']['Accept'] = 'text/event-stream'
        self.new_session()

    def __iter__(self):
        return self

    def __next__(self):
        decoder = codecs.getincrementaldecoder(self.resp.encoding)(errors='replace')
        while not self._event_complete():
            try:
                next_chunk = next(self.resp_iterator)
                if not next_chunk:
                    raise EOFError()
                self.buf += decoder.decode(next_chunk)

            except (StopIteration, requests.RequestException, EOFError):
                time.sleep(self.retry / 1000.0)
                self._run()

                # The SSE spec only supports resuming from a whole message, so
                # if we have half a message we should throw it out.
                head, sep, _ = self.buf.rpartition('\n')
                self.buf = head + sep
                continue

        # Split the complete event (up to the END_OF_FIELD) into event_string,
        # and retain anything after the current complete event in self.buf
        # for next time.
        (event_string, self.buf) = re.split(END_OF_FIELD, self.buf, maxsplit=1)
        msg = Event.parse(event_string)

        # If the server requests a specific retry delay, we need to honor it.
        if msg.retry:
            self.retry = msg.retry

        # last_id should only be set if included in the message.  It's not
        # forgotten if a message omits it.
        if msg.event_id:
            self.last_id = msg.event_id

        if msg.data:
            raw_data = util.json.loads(util.json.loads(msg.data)['data'])['signals']

            data = AccernClient.quant_filter(self.schema, raw_data)
            data = AccernClient.select_fields(self.schema, data)

            if self._listener.on_data(data) is False:
                return False

        return msg

    def _event_complete(self):
        return re.search(END_OF_FIELD, self.buf) is not None

    def _run(self):
        if self.last_id:
            self.requests_kwargs['headers']['Last-Event-ID'] = self.last_id
        # Use session if set.  Otherwise fall back to requests module.
        requester = self.session or requests
        self.resp = requester.get(self.url, stream=True)
        self.resp_iterator = self.resp.iter_content(chunk_size=self.chunk_size)

        # todo: Ensure we're handling redirects.  Might also stick the 'origin'
        # attribute on Events like the Javascript spec requires.
        self.resp.raise_for_status()
        while next(self, None) is not None:
            next(self, None)


    def new_session(self):
        self.session = requests.Session()
        self.session.params = None

    def performs(self):
        """Perform HTTP GET/POST with credentials.

        :param output: output config.

        :param params: HTTP GET parameters.

        :raises ApiError: when the API returns an error.
        :raises Timeout: if the request timed out.
        :raises TransportError: when something went wrong while trying to
            exceute a request.
        """
        print ('%s - Start streaming, use [Ctrl+C] to stop...' % (util.datetime.now()))
        schema = Schema.validate_schema(method='stream', schema=self.schema)
        params = AccernClient.build_api_params(schema)
        params['token'] = AccernClient.check_token(self.token)
        encoded_params = util.urlencode(list(AccernClient.api_encode(params or {})))
        self.url = AccernClient.build_api_url(self.api_base, encoded_params)
        try:
            self._run()
        except KeyboardInterrupt:
            print ('%s - Streaming stopped...' % (util.datetime.now()))
        else:
            pass

    if util.six.PY2:
        next = __next__



"""Accern: python library for Accern API.

Accern is a python library to query, download, filter and save Accern Data.
"""

from accern.api import API
from accern.historical import HistoricalClient
from accern.schema import Schema
from accern.stream import StreamClient, StreamListener
from accern.default_client import AccernClient

token = None

__all__ = [
    '__version__',
    'API',
    'AccernClient',
    'HistoricalClient',
    'Schema',
    'StreamClient',
    'StreamListener'
]



from accern import error, util
from os.path import dirname

MODULE_PATH = dirname(__file__)
FIELD_OPTIONS = util.json.load(open("%s/data/options.json" % MODULE_PATH))

class Schema(object):
    @staticmethod
    def _validate_categorical(field, value):
        VALUE = FIELD_OPTIONS[field]['value']
        return {
            'field': field,
            'type': 'categorical',
            'value': {
                'valid': [v for v in value if v in VALUE],
                'invalid': [v for v in value if v not in VALUE]
            }
        }

    @staticmethod
    def _validate_range(field, value):
        RANGE = FIELD_OPTIONS[field]['value']
        if not isinstance(value, list):
            return {
                'field': field,
                'error': 'Invalid value type of field.'
            }
        if len(value) != 2:
            return {
                'field': field,
                'type': 'range',
                'error': '"%s" has wrong number or arguments.' % (field)
            }
        if value[1] < value[0]:
            return {
                'field': field,
                'type': 'range',
                'error': '"%s" has malformed filter option value.' % (field)
            }
        if value[0] >= RANGE[0] and value[1] <= RANGE[1]:
            return {
                'field': field,
                'type': 'range',
                'value': {
                    'valid': True,
                    'default_range': RANGE
                }
            }
        return {
            'field': field,
            'type': 'range',
            'value': {
                'valid': False,
                'default_range': RANGE
            }
        }

    @staticmethod
    def _validate_norange(field, value):
        if len(value) != 2:
            return {
                'field': field,
                'type': 'no range',
                'error': '"%s" has wrong number or arguments.' % (field)
            }
        if value[1] < value[0]:
            return {
                'field': field,
                'type': 'range',
                'error': '"%s" has malformed filter option value.' % (field)
            }
        return {
            'field': field,
            'type': 'no range'
        }

    @classmethod
    def get_fields(cls):
        return [v for v in sorted(FIELD_OPTIONS.keys()) if 'filter' in FIELD_OPTIONS[v]['method'] or 'url_param' in FIELD_OPTIONS[v]['method']]

    @classmethod
    def get_options(cls, field):
        try:
            if 'filter' in FIELD_OPTIONS[field]['method'] or 'url_param' in FIELD_OPTIONS[field]['method']:
                options = FIELD_OPTIONS[field]
                del options['method']
                return options
        except KeyError:
            raise error.SchemaError('Invalid field (%s) in filter option.' % field)

    @classmethod
    def get_url_params(cls):
        return [k for k, v in FIELD_OPTIONS.items() if 'url_param' in v['method']]

    @classmethod
    def validate_options(cls, field=None, value=None):
        if field not in FIELD_OPTIONS:
            raise error.SchemaError('Invalid field (%s) in filter option.' % field)
        if value is None:
            raise error.SchemaError('No filter option value for "%s".' % field)

        types = {
            'categorical': cls._validate_categorical,
            'norange': cls._validate_norange,
            'range': cls._validate_range,
            'other': lambda field, value: {'field': field, 'value': value}
        }
        return types[FIELD_OPTIONS[field]['type']](field, value)

    @classmethod
    def validate_schema_filters(cls, method, filters):
        if isinstance(filters, list):
            if method != 'historical':
                raise error.SchemaError('Method "%s" does not support multiple filters.' % method)
            for f in filters:
                cls.validate_schema_filters(method=method, filters=f)
            return
        for f in filters:
            if isinstance(filters[f], list) and any(isinstance(el, list) for el in filters[f]):
                for el in filters[f]:
                    resp = cls.validate_options(field=f, value=el)
                    if 'error' in resp:
                        raise error.SchemaError(resp['error'])
            else:
                resp = cls.validate_options(field=f, value=filters[f])
                if 'error' in resp:
                    raise error.SchemaError(resp['error'])

    @classmethod
    def validate_schema_select(cls, method, select):
        if not select:
            return None

        if method in ['api', 'stream']:
            if any('function' in el for el in select):
                raise error.SchemaError('Method "%s" does not support select field functions.' % method)
            if isinstance(select, list):
                if any('field' not in v for v in select):
                    raise error.SchemaError('Missing "field" in select option.')
                return [{'field': v['field'], 'alias': v.get('alias', v['field'])} for v in select]

            else:
                if 'field' not in select:
                    raise error.SchemaError('Missing "field" in select option.')
                return {'field': select['field'], 'alias': select.get('alias', select['field'])}
        else:
            try:
                if any(v["field"] == 'harvested_at' and v.get("function") is not None for v in select):
                    select = [{
                        'field': v['field'],
                        'alias': v.get('alias', v['field']),
                        'function': v.get('function', FIELD_OPTIONS[v['field']]['function'][0])
                    } for v in select]

                    if any(v['field'] == 'harvested_at' and v['alias'] != v['function'] for v in select):
                        raise error.SchemaError("Alias of harvested_at is different from it's aggregation function.")
                else:
                    select = [
                        {
                            'field': v['field'],
                            'alias': v.get('alias', v['field'])
                        } for v in select]
                return select
            except KeyError:
                raise error.SchemaError('Missing "field" in select option.')

    @classmethod
    def validate_schema(cls, method=None, schema=None):
        if schema is None:
            return None
        schema = {k.lower(): v for k, v in schema.items()}

        if method is None:
            raise error.SchemaError('Method is missing.')

        if method not in ['api', 'historical', 'stream']:
            raise error.SchemaError('Illegal usage of validate schema function.')

        if method in ['api', 'stream']:
            if 'name' in schema:
                raise error.SchemaError('Illegal "name" in %s schema.' % method)
            if 'description' in schema:
                raise error.SchemaError('Illegal "description" in %s schema.' % method)
        elif method == 'historical':
            if 'name' not in schema:
                raise error.SchemaError('Required field "name" not found in %s schema.' % method)
            if 'description' not in schema:
                raise error.SchemaError('Required field "description" not found in %s schema.' % method)

        if method in ['api', 'historical', 'stream']:
            filters = schema.get('filters', {})
            cls.validate_schema_filters(method=method, filters=filters)
            select = schema.get('select', {})
            select = cls.validate_schema_select(method=method, select=select)
            if select is not None:
                schema['select'] = select

        return schema



from accern import error, util
from accern.schema import Schema
from datetime import datetime
import re
import sys
import textwrap

try:
    import requests
except ImportError:
    requests = None
else:
    try:
        # Require version 0.8.8, but don't want to depend on distutils
        version = requests.__version__
        major, minor, patch = [int(i) for i in version.split('.')]
    except Exception:
        # Probably some new-fangled version, so it should support verify
        pass
    else:
        if (major, minor, patch) < (0, 8, 8):
            sys.stderr.write(
                'Warning: the Accern library requires that your Python '
                '"requests" library be newer than version 0.8.8, but your '
                '"requests" library is version %s. Accern will fall back to '
                'an alternate HTTP library so everything should work. We '
                'recommend upgrading your "requests" library. If you have any '
                'questions, please contact support@accern.com. (HINT: running '
                '"pip install -U requests" should upgrade your requests '
                'library to the latest version.)' % (version,))
            requests = None


__all__ = [
    'AccernClient',
    'Event',
    'new_http_client'
]


QUANT = [
    'event_impact_gt_mu_add_sigma',
    'event_impact_gt_mu_pos_add_sigma_pos',
    'event_impact_gt_mu_pos_add_2sigma_pos',
    'event_impact_gt_1pct_pos',
    'event_impact_lt_mu_sub_sigma',
    'event_impact_lt_mu_neg_sub_sigma_neg',
    'event_impact_lt_mu_neg_sub_2sigma_neg',
    'event_impact_lt_1pct_neg',
    'event_impact_neg',
    'event_impact_pct_change_avg',
    'event_impact_pct_change_stdev',
    'event_impact_pos',
    'event_relevance',
    'event_sentiment',
    'event_source_timeliness_score',
    'entity_exchange',
    'entity_relevance',
    'entity_sentiment',
    'harvested_at',
    'story_group_sentiment_avg',
    'story_group_sentiment_stdev',
    'story_group_count',
    'story_group_traffic_sum',
    'story_sentiment',
    'story_traffic'
]


def new_http_client(*args, **kwargs):
    return RequestsClient(*args, **kwargs)


class AccernClient(object):
    @staticmethod
    def api_encode(params):
        if isinstance(params, object):
            for key, value in util.six.iteritems(params):
                key = util.utf8(key)
                if value is None:
                    continue
                elif isinstance(value, list):
                    yield (key, ",".join(value))
                else:
                    yield (key, util.utf8(value))

    @staticmethod
    def build_api_url(url, query):
        scheme, netloc, path, base_query, fragment = util.urlsplit(url)
        if base_query:
            query = '%s&%s' % (base_query, query)
        return util.urlunsplit((scheme, netloc, path, query, fragment))

    @staticmethod
    def build_api_headers(token, method):
        if method == "POST":
            return {
                'Content-Type': 'application/json',
                'IO-Authorization': token
            }
        elif method == "GET":
            return {
                'Content-Type': 'application/json',
                'IO-Authorization': token
            }

    @staticmethod
    def build_api_params(schema):
        if schema is None:
            filters = {}
        else:
            filters = schema.get('filters', {})
        avail_params = [value for value in filters if value in Schema.get_url_params()]
        params = {key: filters[key] for key in avail_params}
        if 'harvested_at' in filters:
            del params['harvested_at']
            params['from'] = filters['harvested_at'][0]
        return params

    @classmethod
    def check_values(cls, raw_data, f, f_values):
        if f == 'harvested_at':
            for data in raw_data:
                if (datetime.strptime(data[f], '%Y-%m-%dT%H:%M:%S.%fZ') >= datetime.strptime(f_values[0], '%Y-%m-%d %H:%M:%S')) and \
                   (datetime.strptime(data[f], '%Y-%m-%dT%H:%M:%S.%fZ') <= datetime.strptime(f_values[1], '%Y-%m-%d %H:%M:%S')):
                    yield data
        else:
            for data in raw_data:
                for value in f_values:
                    if data[f] >= value[0] and data[f] <= value[1]:
                        yield data

    @staticmethod
    def check_token(token):
        if token:
            my_token = token
        else:
            from accern import token
            my_token = token

        if my_token is None:
            raise error.AuthenticationError('No token provided.')
        return my_token

    @staticmethod
    def handle_error(rbody, rcode, resp):
        try:
            error_data = resp['error']
        except (KeyError, TypeError):
            raise error.APIError(
                "Invalid response object from API: %r (HTTP response code "
                "was %d)" % (rbody, rcode), rbody, rcode, resp)

        if rcode == 400:
            return error.AccernError(error_data, rbody, rcode)

    @staticmethod
    def select_fields(schema, raw_data):
        if schema is None:
            select = []
        else:
            select = schema.get('select', [])
        names = [option['alias'] if 'alias' in option else option['field'] for option in select]
        fields = [option['field'] for option in select]
        data_selected = []

        for data in raw_data:
            if bool(select) > 0:
                try:
                    if isinstance(select, list):
                        new_data = dict(zip(names, [data[field] for field in fields]))
                    else:
                        raise error.AccernError('Select field should be a list.')
                except KeyError:
                    raise error.AccernError('Invalid select values passed.')
            else:
                new_data = data

            if 'entity_competitors' in new_data:
                new_data['entity_competitors'] = ' | '.join(new_data['entity_competitors'])
            if 'entity_indices' in new_data:
                new_data['entity_indices'] = ' | '.join(new_data['entity_indices'])

            data_selected.append(new_data)

        return data_selected

    @classmethod
    def quant_filter(cls, schema, raw_data):
        if schema is None:
            filters = {}
        else:
            filters = schema.get('filters', {})

        for f in filters:
            if f in QUANT:
                data_filtered = list(cls.check_values(raw_data, f, filters[f]))
                raw_data = data_filtered

        return raw_data


class Event(object):
    SSE_LINE_PATTERN = re.compile('(?P<name>[^:]*):?( ?(?P<value>.*))?')

    def __init__(self, data='', event='message', event_id=None, retry=None):
        self.data = data
        self.event = event
        self.event_id = event_id
        self.retry = retry

    def dump(self):
        lines = []
        if self.event_id:
            lines.append('id: %s' % self.event_id)

        # Only include an event line if it's not the default already.
        if self.event != 'message':
            lines.append('event: %s' % self.event)

        if self.retry:
            lines.append('retry: %s' % self.retry)

        lines.extend('data: %s' % d for d in self.data.split('\n'))
        return '\n'.join(lines) + '\n\n'

    @classmethod
    def parse(cls, raw):
        """Given a possibly-multiline string representing an SSE message, parse it and return a Event object."""
        msg = cls()
        for line in raw.splitlines():
            m = cls.SSE_LINE_PATTERN.match(line)
            if m is None:
                # Malformed line.  Discard but warn.
                sys.stderr.write('Invalid SSE line: "%s"' % line, SyntaxWarning)
                continue

            name = m.group('name')
            if name == '':
                # line began with a ":", so is a comment.  Ignore
                continue
            value = m.group('value')

            if name == 'data':
                # If we already have some data, then join to it with a newline.
                # Else this is it.
                if msg.data:
                    msg.data = '%s\n%s' % (msg.data, value)
                else:
                    msg.data = value
            elif name == 'event':
                msg.event = value
            elif name == 'id':
                msg.event_id = value
            elif name == 'retry':
                msg.retry = int(value)

        return msg

    def __str__(self):
        return self.data


class HTTPClient(object):
    def request(self, method, url, headers, post_data=None):
        raise NotImplementedError(
            'HTTPClient subclasses must implement `request`')


class RequestsClient(HTTPClient):
    name = 'requests'

    def __init__(self, timeout=80, session=None, **kwargs):
        super(RequestsClient, self).__init__(**kwargs)
        self._timeout = timeout
        self._session = session or requests.Session()

    def request(self, method, url, headers, post_data=None):
        kwargs = {}
        try:
            try:
                result = self._session.request(method,
                                               url,
                                               headers=headers,
                                               data=post_data,
                                               timeout=self._timeout,
                                               **kwargs)
            except TypeError as e:
                raise TypeError(
                    'Warning: It looks like your "requests" library is out of '
                    'date. You can fix that by running "pip install -U requests".) '
                    'The underlying error was: %s' % (e))
            content = result.content
            status_code = result.status_code
        except Exception as e:
            # Would catch just requests.exceptions.RequestException, but can
            # also raise ValueError, RuntimeError, etc.
            self.handle_request_error(e)
        return content, status_code, result.headers

    @staticmethod
    def handle_request_error(e):
        if isinstance(e, requests.exceptions.RequestException):
            msg = ("Unexpected error communicating with Accern.  "
                   "If this problem persists, let us know at "
                   "support@accern.com.")
            err = "%s: %s" % (type(e).__name__, str(e))
        else:
            msg = ("Unexpected error communicating with Accern. "
                   "It looks like there's probably a configuration "
                   "issue locally.  If this problem persists, let us "
                   "know at support@accern.com.")
            err = "A %s was raised" % (type(e).__name__,)
            if str(e):
                err += " with error message %s" % (str(e),)
            else:
                err += " with no error message"
        msg = textwrap.fill(msg) + "\n\n(Network error: %s)" % (err,)
        raise error.APIConnectionError(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'accern'))
from version import __version__

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.version_info <= (2, 7):
    error = 'Requires Python Version 2.7 or above... exiting.'
    print (sys.stderr, error)
    sys.exit(1)

__long_description__ = """
*Accern*: A Python library for Accern's API.

Accern primarily serves institutional investors. The majority of our current
clients are quantitative hedge funds. Many small firms are coming to us
because of our flexible and affordable pricing options. Large firms are coming
to us due to our dedicated support, news source customization, and much more.
Aside from hedge funds, our existing clients include pension and endowment
funds, banks, brokerage firms, and more.

Accern library for Python helps user to get fast, flexible data structures from
Accern historical and streaming api.

All Accern wheels from PyPI are MIT licensed.

"""

__requirements__ = [
    'requests>=2.18.4',
    'six>=1.10.0'
]

setup(
    name='Accern',
    version=__version__,
    description="A python library for Accern Data API",
    long_description=__long_description__,
    license='MIT License',
    url="https://github.com/Accern/accern-python",
    packages=['accern'],
    package_data={"accern": ['data/*.json']},
    platforms='Posix; MacOS X; Windows',
    setup_requires=__requirements__,
    include_package_data=True,
    install_requires=__requirements__,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'Topic :: Office/Business :: Financial :: Investment',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
    ]
)
